@@ -10,12 +10,14 @@
 #   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 #   License for the specific language governing permissions and limitations
 #   under the License.
-#
 
-import munch
+import functools
+import ipaddress
+
 from openstackclient.identity import common as identity_common
 from osc_lib import exceptions as osc_exc
 from osc_lib import utils
+from oslo_utils import uuidutils
 
 from octaviaclient.api import exceptions
 from octaviaclient.osc.v2 import constants
@@ -62,6 +64,62 @@ def _map_attrs(args, source_attr_map):
     return res
 
 
+def _find_resource(list_funct, resource_name, root_tag, name, parent=None):
+    """Search for a resource by name and ID.
+
+    This function will search for a resource by both the name and ID,
+    returning the resource once it finds a match. If no match is found,
+    an exception will be raised.
+
+    :param list_funct: The resource list method to call during searches.
+    :param resource_name: The name of the resource type we are searching for.
+    :param root_tag: The root tag of the resource returned from the API.
+    :param name: The value we are searching for, a resource name or ID.
+    :param parent: The parent resource ID, when required.
+    :return: The resource found for the name or ID.
+    :raises osc_exc.CommandError: If more than one match or none are found.
+    """
+    if parent:
+        parent_args = [parent]
+    else:
+        parent_args = []
+    # Optimize the API call order if we got a UUID-like name or not
+    if uuidutils.is_uuid_like(name):
+        # Try by ID first
+        resource = list_funct(*parent_args, id=name)[root_tag]
+        if len(resource) == 1:
+            return resource[0]
+
+        # Try by name next
+        resource = list_funct(*parent_args, name=name)[root_tag]
+        if len(resource) == 1:
+            return resource[0]
+        if len(resource) > 1:
+            msg = ("{0} {1} found with name or ID of {2}. Please try "
+                   "again with UUID".format(len(resource), resource_name,
+                                            name))
+            raise osc_exc.CommandError(msg)
+    else:
+        # Try by name first
+        resource = list_funct(*parent_args, name=name)[root_tag]
+        if len(resource) == 1:
+            return resource[0]
+        if len(resource) > 1:
+            msg = ("{0} {1} found with name or ID of {2}. Please try "
+                   "again with UUID".format(len(resource), resource_name,
+                                            name))
+            raise osc_exc.CommandError(msg)
+
+        # Try by ID next
+        resource = list_funct(*parent_args, id=name)[root_tag]
+        if len(resource) == 1:
+            return resource[0]
+
+    # We didn't find what we were looking for, raise a consistent error.
+    msg = "Unable to locate {0} in {1}".format(name, resource_name)
+    raise osc_exc.CommandError(msg)
+
+
 def get_resource_id(resource, resource_name, name):
     """Converts a resource name into a UUID for consumption for the API
 
@@ -94,34 +152,65 @@ def get_resource_id(resource, resource_name, name):
                 ).id
                 return project_id
             return 'non-uuid'
-        if resource_name == 'members':
-            names = [re for re in resource(name['pool_id'])['members']
-                     if re.get('id') == name['member_id'] or
-                     re.get('name') == name['member_id']]
-            name = name['member_id']
-            if len(names) > 1:
-                msg = ("{0} {1} found with name or ID of {2}. Please try "
-                       "again with UUID".format(len(names), resource_name,
-                                                name))
-                raise osc_exc.CommandError(msg)
-            return names[0].get('id')
-        if resource_name == 'l7rules':
-            names = [re for re in resource(name['l7policy_id'])['rules']
-                     if re.get('id') == name['l7rule_id']]
-            name = name['l7rule_id']
-            return names[0].get('id')
-        names = [re for re in resource()[resource_name]
-                 if re.get('name') == name or re.get('id') == name]
-        if len(names) > 1:
-            msg = ("{0} {1} found with name or ID of {2}. Please try "
-                   "again with UUID".format(len(names), resource_name,
-                                            name))
-            raise osc_exc.CommandError(msg)
-        return names[0].get(primary_key)
 
-    except IndexError:
+        if resource_name == 'members':
+            member = _find_resource(resource, resource_name, 'members',
+                                    name['member_id'], parent=name['pool_id'])
+            return member.get('id')
+
+        if resource_name == 'l7rules':
+            l7rule = _find_resource(resource, resource_name, 'rules',
+                                    name['l7rule_id'],
+                                    parent=name['l7policy_id'])
+            return l7rule.get('id')
+
+        resource = _find_resource(resource, resource_name, resource_name, name)
+        return resource.get(primary_key)
+
+    except IndexError as e:
         msg = "Unable to locate {0} in {1}".format(name, resource_name)
-        raise osc_exc.CommandError(msg)
+        raise osc_exc.CommandError(msg) from e
+
+
+def add_tags_attr_map(attr_map):
+    tags_attr_map = {
+        'tags': ('tags', list),
+        'any_tags': ('tags-any', list),
+        'not_tags': ('not-tags', list),
+        'not_any_tags': ('not-tags-any', list),
+    }
+    attr_map.update(tags_attr_map)
+
+
+def validate_vip_dict(vip_dict, client_manager):
+    # We have validation in two places -- _map_attrs checks sub-resources, and
+    # later _check_attrs does further api-specific validation. We need both for
+    # additional vips, so we may as well just do both here while we're at it.
+    if 'subnet_id' not in vip_dict:
+        raise osc_exc.CommandError(
+            'Additional VIPs must include a subnet-id.')
+    subnet_id = get_resource_id(client_manager.neutronclient.list_subnets,
+                                'subnets', vip_dict['subnet_id'])
+    vip_dict['subnet_id'] = subnet_id
+    if 'ip_address' in vip_dict:
+        try:
+            ipaddress.ip_address(vip_dict['ip_address'])
+        except ValueError as e:
+            raise osc_exc.CommandError(str(e))
+
+
+def handle_additional_vips(vips, client_manager):
+    additional_vips = []
+    for vip in vips:
+        vip_dict = {}
+        parts = vip.split(',')
+        for part in parts:
+            k, v = part.split('=')
+            vip_dict[k.replace('-', '_')] = v
+        validate_vip_dict(vip_dict, client_manager)
+        additional_vips.append(vip_dict)
+
+    return additional_vips
 
 
 def get_loadbalancer_attrs(client_manager, parsed_args):
@@ -174,8 +263,13 @@ def get_loadbalancer_attrs(client_manager, parsed_args):
             client_manager.load_balancer.flavor_list
         ),
         'availability_zone': ('availability_zone', str),
-
+        'additional_vip': (
+            'additional_vips',
+            functools.partial(
+                handle_additional_vips, client_manager=client_manager)
+        ),
     }
+    add_tags_attr_map(attr_map)
 
     _attrs = vars(parsed_args)
     attrs = _map_attrs(_attrs, attr_map)
@@ -226,7 +320,10 @@ def get_listener_attrs(client_manager, parsed_args):
                                      _format_str_if_need_treat_unset),
         'allowed_cidrs': ('allowed_cidrs', list),
         'tls_ciphers': ('tls_ciphers', str),
+        'tls_versions': ('tls_versions', list),
+        'alpn_protocols': ('alpn_protocols', list),
     }
+    add_tags_attr_map(attr_map)
 
     _attrs = vars(parsed_args)
     attrs = _map_attrs(_attrs, attr_map)
@@ -273,7 +370,10 @@ def get_pool_attrs(client_manager, parsed_args):
         'enable_tls': ('tls_enabled', lambda x: True),
         'disable_tls': ('tls_enabled', lambda x: False),
         'tls_ciphers': ('tls_ciphers', str),
+        'tls_versions': ('tls_versions', list),
+        'alpn_protocols': ('alpn_protocols', list),
     }
+    add_tags_attr_map(attr_map)
 
     _attrs = vars(parsed_args)
     attrs = _map_attrs(_attrs, attr_map)
@@ -315,6 +415,7 @@ def get_member_attrs(client_manager, parsed_args):
         'enable': ('admin_state_up', lambda x: True),
         'disable': ('admin_state_up', lambda x: False),
     }
+    add_tags_attr_map(attr_map)
 
     _attrs = vars(parsed_args)
     attrs = _map_attrs(_attrs, attr_map)
@@ -354,6 +455,7 @@ def get_l7policy_attrs(client_manager, parsed_args):
         'enable': ('admin_state_up', lambda x: True),
         'disable': ('admin_state_up', lambda x: False)
     }
+    add_tags_attr_map(attr_map)
 
     _attrs = vars(parsed_args)
     attrs = _map_attrs(_attrs, attr_map)
@@ -388,6 +490,7 @@ def get_l7rule_attrs(client_manager, parsed_args):
         'enable': ('admin_state_up', lambda x: True),
         'disable': ('admin_state_up', lambda x: False)
     }
+    add_tags_attr_map(attr_map)
 
     _attrs = vars(parsed_args)
     attrs = _map_attrs(_attrs, attr_map)
@@ -426,6 +529,7 @@ def get_health_monitor_attrs(client_manager, parsed_args):
         'http_version': ('http_version', float),
         'domain_name': ('domain_name', str)
     }
+    add_tags_attr_map(attr_map)
 
     _attrs = vars(parsed_args)
     attrs = _map_attrs(_attrs, attr_map)
@@ -440,6 +544,8 @@ def get_quota_attrs(client_manager, parsed_args):
         'load_balancer': ('load_balancer', int),
         'member': ('member', int),
         'pool': ('pool', int),
+        'l7policy': ('l7policy', int),
+        'l7rule': ('l7rule', int),
         'project': (
             'project_id',
             'project',
@@ -599,13 +705,22 @@ def _format_str_if_need_treat_unset(data):
 
 
 def get_unsets(parsed_args):
-    return {arg: None for arg, value in vars(parsed_args).items() if
-            value is True and arg != 'wait'}
+    unsets = {}
+    for arg, value in vars(parsed_args).items():
+        if value and arg == 'tags':
+            unsets[arg] = value
+        elif value is True and arg not in ('wait', 'all_tag'):
+            unsets[arg] = None
+    return unsets
+
+
+class _Munch(dict):
+    __getattr__ = dict.get
 
 
 def wait_for_active(status_f, res_id):
     success = utils.wait_for_status(
-        status_f=lambda x: munch.Munch(status_f(x)),
+        status_f=lambda x: _Munch(status_f(x)),
         res_id=res_id,
         status_field=constants.PROVISIONING_STATUS,
         sleep_time=3
@@ -616,17 +731,18 @@ def wait_for_active(status_f, res_id):
             message="The resource did not successfully reach ACTIVE status.")
 
 
-def wait_for_delete(status_f, res_id):
+def wait_for_delete(status_f, res_id,
+                    status_field=constants.PROVISIONING_STATUS):
     class Getter(object):
         @staticmethod
         def get(id):
-            return munch.Munch(status_f(id))
+            return _Munch(status_f(id))
 
     try:
         success = utils.wait_for_delete(
             manager=Getter,
             res_id=res_id,
-            status_field=constants.PROVISIONING_STATUS,
+            status_field=status_field,
             sleep_time=3
         )
         if not success:
@@ -636,3 +752,23 @@ def wait_for_delete(status_f, res_id):
     except exceptions.OctaviaClientException as e:
         if e.code != 404:
             raise
+
+
+def set_tags_for_set(resource_get, resource_id, attrs, clear_tags=False):
+    if attrs.get('tags'):
+        resource = resource_get(resource_id)
+        tags = set([] if clear_tags else resource['tags'])
+        tags |= set(attrs['tags'])
+        attrs['tags'] = list(tags)
+    elif clear_tags:
+        attrs['tags'] = []
+
+
+def set_tags_for_unset(resource_get, resource_id, attrs, clear_tags=False):
+    if clear_tags:
+        attrs['tags'] = []
+    elif attrs.get('tags'):
+        resource = resource_get(resource_id)
+        tags = set(resource['tags'])
+        tags -= set(attrs['tags'])
+        attrs['tags'] = list(tags)
